from .BasAdapter import BasAdapter
from .ChdAdapter import ChdAdapter
from .DisAdapter import DisAdapter
from .DrnAdapter import DrnAdapter
from .EvtAdapter import EvtAdapter
from .FhbAdapter import FhbAdapter
from .GhbAdapter import GhbAdapter
from .HobAdapter import HobAdapter
from .LmtAdapter import LmtAdapter
from .LpfAdapter import LpfAdapter
from .MfAdapter import MfAdapter
from .NwtAdapter import NwtAdapter
from .OcAdapter import OcAdapter
from .PcgAdapter import PcgAdapter
from .RchAdapter import RchAdapter
from .RivAdapter import RivAdapter
from .UpwAdapter import UpwAdapter
from .WelAdapter import WelAdapter




from .AdvAdapter import AdvAdapter
from .BtnAdapter import BtnAdapter
from .DspAdapter import DspAdapter
from .GcgAdapter import GcgAdapter
from .LktAdapter import LktAdapter
from .MtAdapter import MtAdapter
from .PhcAdapter import PhcAdapter
from .RctAdapter import RctAdapter
from .SftAdapter import SftAdapter
from .SsmAdapter import SsmAdapter
from .TobAdapter import TobAdapter
from .UztAdapter import UztAdapter



import flopy.mt3d as mt


class UztAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mt):
        content = self.merge()
        return mt.Mt3dUzt(
            _mt,
            **content
        )

    @staticmethod
    def default():
        default = {
            "mxuzcon": 0,
            "icbcuz": None,
            "iet": 0,
            "iuzfbnd": None,
            "wc": 0.0,
            "sdh": 0.0,
            "cuzinf": None,
            "cuzet": None,
            "cgwet": None,
            "extension": 'uzt',
            "unitnumber": None,
            "filenames": None
        }
        return default

    @staticmethod
    def read_package(package):
        content = {
            "mxuzcon": package.mxuzcon,
            "icbcuz": package.icbcuz,
            "iet": package.iet,
            "iuzfbnd": package.iuzfbnd,
            "wc": package.wc,
            "sdh": package.sdh,
            "cuzinf": package.cuzinf,
            "cuzet": package.cuzet,
            "cgwet": package.cgwet,
            "extension": package.extension,
            "unitnumber": package.unitnumber,
            "filenames": package.filenames
        }
        return content



import flopy.modpath as mp


class SimAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mp):
        content = self.merge()
        return mp.ModpathSim(
            _mp,
            **content
        )

    @staticmethod
    def default():
        return {
            "mp_name_file": 'mp.nam',
            "mp_list_file": 'mp.list',
            "option_flags": [1, 2, 1, 1, 1, 2, 2, 1, 2, 1, 1, 1],
            "ref_time": 0,
            "ref_time_per_stp": [0, 0, 1.0],
            "stop_time": None,
            "group_name": ['group_1'],
            "group_placement": [[1, 1, 1, 0, 1, 1]],
            "release_times": [[1, 1]],
            "group_region": [[1, 1, 1, 1, 1, 1]],
            "mask_nlay": [1],
            "mask_layer": [1],
            "mask_1lay": [1],
            "face_ct": [1],
            "ifaces": [[6, 1, 1]],
            "part_ct": [[1, 1, 1]],
            "time_ct": 1,
            "release_time_incr": 1,
            "time_pts": [1],
            "particle_cell_cnt": [[2, 2, 2]],
            "cell_bd_ct": 1,
            "bud_loc": [[1, 1, 1, 1]],
            "trace_id": 1,
            "stop_zone": 1,
            "zone": 1,
            "retard_fac": 1.0,
            "retard_fcCB": 1.0,
            "strt_file": None,
            "extension": 'mpsim'
        }

    @staticmethod
    def read_package(package):
        return {
            "mp_name_file": package.mp_name_file,
            "mp_list_file": package.mp_list_file,
            "option_flags": package.option_flags,
            "ref_time": package.ref_time,
            "ref_time_per_stp": package.ref_time_per_stp,
            "stop_time": package.stop_time,
            "group_name": package.group_name,
            "group_placement": package.group_placement,
            "release_times": package.release_times,
            "group_region": package.group_region,
            "mask_nlay": package.mask_nlay,
            "mask_layer": package.mask_layer,
            "mask_1lay": package.mask_1lay,
            "face_ct": package.face_ct,
            "ifaces": package.ifaces,
            "part_ct": package.part_ct,
            "time_ct": package.time_ct,
            "release_time_incr": package.release_time_incr,
            "time_pts": package.time_pts,
            "particle_cell_cnt": package.particle_cell_cnt,
            "cell_bd_ct": package.cell_bd_ct,
            "bud_loc": package.bud_loc,
            "trace_id": package.trace_id,
            "stop_zone": package.stop_zone,
            "zone": package.zone,
            "retard_fac": package.retard_fac,
            "retard_fcCB": package.retard_fcCB,
            "strt_file": package.strt_file,
            "extension": package.extension
        }



import flopy.modflow as mf


class RchAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mf):
        content = self.merge()
        return mf.ModflowRch(
            _mf,
            **content
        )

    @staticmethod
    def default():
        return {
            "nrchop": 3,
            "ipakcb": 0,
            "rech": 0,
            "irch": 0,
            "extension": 'rch',
            "unitnumber": 19
        }

    @staticmethod
    def read_package(package):
        return {
            "nrchop": package.nrchop,
            "ipakcb": package.ipakcb,
            "rech": package.rech,
            "irch": package.irch,
            "extension": package.extension,
            "unitnumber": package.unitnumber
        }



import flopy.modflow as mf


class UpwAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mf):
        content = self.merge()
        return mf.ModflowUpw(
            _mf,
            **content
        )

    @staticmethod
    def default():
        default = {
            "laytyp": 0,
            "layavg": 0,
            "chani": 1.0,
            "layvka": 0,
            "laywet": 0,
            "ipakcb": 53,
            "hdry": -1e+30,
            "iphdry": 0,
            "hk": 1.0,
            "hani": 1.0,
            "vka": 1.0,
            "ss": 1e-5,
            "sy": 0.15,
            "vkcb": 0.0,
            "extension": 'upw',
            "unitnumber": 31
        }

        return default

    @staticmethod
    def read_package(package):
        content = {
            "laytyp": package.laytyp.array.tolist(),
            "layavg": package.layavg.array.tolist(),
            "chani": package.chani.array.tolist(),
            "layvka": package.layvka.array.tolist(),
            "laywet": package.laywet.array.tolist(),
            "ipakcb": package.ipakcb,
            "hdry": package.hdry,
            "iphdry": package.iphdry,
            "hk": package.hk.array.tolist(),
            "hani": package.hani.array.tolist(),
            "vka": package.vka.array.tolist(),
            "ss": package.ss.array.tolist(),
            "sy": package.sy.array.tolist(),
            "vkcb": package.vkcb.array.tolist(),
            "extension": package.extension[0],
            "unitnumber": package.unit_number[0]
        }
        return content



import flopy.modflow as mf


class ChdAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            if key == 'stress_period_data':
                default[key] = self.to_dict(self._data[key])
                continue

            default[key] = self._data[key]
        return default

    def to_dict(self, data):
        if type(data) == list:
            spd_dict = {}
            for stress_period, record in enumerate(data):
                spd_dict[stress_period] = record
            return spd_dict
        return data

    def get_package(self, _mf):
        content = self.merge()
        return mf.ModflowChd(
            _mf,
            **content
        )

    @staticmethod
    def default():
        default = {
            "stress_period_data": None,
            "dtype": None,
            "extension": 'chd',
            "unitnumber": 24
        }

        return default

    @staticmethod
    def read_package(package):
        content = {
            # stress period data values translated to list of lists to be json serializable
            "stress_period_data": {k: [list(i) for i in v] for k, v in package.stress_period_data.data.items()},
            # "dtype": package.dtype,
            "extension": package.extension[0],
            "unitnumber": package.unit_number[0]
        }
        return content



import numpy as np
import flopy.mt3d as mt


class BtnAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            if not key.startswith('_'):
                default[key] = self._data[key]
        return default

    def get_package(self, _mt):
        content = self.merge()
        return mt.Mt3dBtn(
            _mt,
            **content
        )

    @staticmethod
    def default():
        default = {
            "MFStyleArr": False,
            "DRYCell": False,
            "Legacy99Stor": False,
            "FTLPrint": False,
            "NoWetDryPrint": False,
            "OmitDryBud": False,
            "AltWTSorb": False,
            "nlay": None,
            "nrow": None,
            "ncol": None,
            "nper": None,
            "ncomp": 1,
            "mcomp": 1,
            "tunit": 'D',
            "lunit": 'M',
            "munit": 'KG',
            "laycon": None,
            "delr": None,
            "delc": None,
            "htop": None,
            "dz": None,
            "prsity": 0.3,
            "icbund": 1,
            "sconc": 0.0,
            "cinact": 1e+30,
            "thkmin": 0.01,
            "ifmtcn": 0,
            "ifmtnp": 0,
            "ifmtrf": 0,
            "ifmtdp": 0,
            "savucn": True,
            "nprs": 0,
            "timprs": None,
            "obs": None,
            "nprobs": 1,
            "chkmas": True,
            "nprmas": 1,
            "perlen": None,
            "nstp": None,
            "tsmult": None,
            "ssflag": None,
            "dt0": 0,
            "mxstrn": 50000,
            "ttsmult": 1.0,
            "ttsmax": 0,
            "species_names": None,
            "extension": 'btn',
            "unitnumber": None,
            "filenames": None
        }
        return default

    @staticmethod
    def read_package(package):
        content = {
            "MFStyleArr": package.MFStyleArr,
            "DRYCell": package.DRYCell,
            "Legacy99Stor": package.Legacy99Stor,
            "FTLPrint": package.FTLPrint,
            "NoWetDryPrint": package.NoWetDryPrint,
            "OmitDryBud": package.OmitDryBud,
            "AltWTSorb": package.AltWTSorb,
            "nlay": package.nlay,
            "nrow": package.nrow,
            "ncol": package.ncol,
            "nper": package.nper,
            "ncomp": package.ncomp,
            "mcomp": package.mcomp,
            "tunit": package.tunit,
            "lunit": package.lunit,
            "munit": package.munit,
            "laycon": package.laycon.array.tolist(),
            "delr": np.reshape(package.delr.array, (len(package.delr.array),)).tolist(),
            "delc": np.reshape(package.delc.array, (len(package.delc.array),)).tolist(),
            "htop": package.htop.array.tolist(),
            "dz": package.dz.array.tolist(),
            "prsity": package.prsity.array.tolist(),
            "icbund": package.icbund.array.tolist(),
            "sconc": [i.array.tolist() for i in package.sconc],
            "cinact": package.cinact,
            "thkmin": package.thkmin,
            "ifmtcn": package.ifmtcn,
            "ifmtnp": package.ifmtnp,
            "ifmtrf": package.ifmtrf,
            "ifmtdp": package.ifmtdp,
            "savucn": package.savucn,
            "nprs": package.nprs,
            "timprs": package.timprs,  # None
            "obs": package.obs,  # None
            "nprobs": package.nprobs,
            "chkmas": package.chkmas,
            "nprmas": package.nprmas,
            "perlen": package.perlen.array.tolist(),
            "nstp": package.nstp.array.tolist(),
            "tsmult": package.tsmult.array.tolist(),
            # "ssflag": package.ssflag,
            "dt0": package.dt0.array.tolist(),
            "mxstrn": package.mxstrn.array.tolist(),
            "ttsmult": package.ttsmult.array.tolist(),
            "ttsmax": package.ttsmax.array.tolist(),
            "species_names": package.species_names if package.species_names else None,
            "extension": package.extension[0],
            "unitnumber": package.unit_number[0],
            # "filenames": package.filenames
        }
        return content



"""
This module is an intermediate layer between flopy version 3.2
and the inowas-modflow-configuration format.

Serialize Modflow and MT3D model into JSON format
"""
import os
import json
import flopy.modflow
import flopy.mt3d
from FlopyAdapter.MfPackages.BasAdapter import BasAdapter
from FlopyAdapter.MfPackages.ChdAdapter import ChdAdapter
from FlopyAdapter.MfPackages.DisAdapter import DisAdapter
from FlopyAdapter.MfPackages.LpfAdapter import LpfAdapter
from FlopyAdapter.MfPackages.MfAdapter import MfAdapter
from FlopyAdapter.MfPackages.NwtAdapter import NwtAdapter
from FlopyAdapter.MfPackages.PcgAdapter import PcgAdapter
from FlopyAdapter.MfPackages.WelAdapter import WelAdapter
from FlopyAdapter.MfPackages.LmtAdapter import LmtAdapter
from FlopyAdapter.MtPackages.MtAdapter import MtAdapter
from FlopyAdapter.MtPackages.AdvAdapter import AdvAdapter
from FlopyAdapter.MtPackages.BtnAdapter import BtnAdapter
from FlopyAdapter.MtPackages.DspAdapter import DspAdapter
from FlopyAdapter.MtPackages.GcgAdapter import GcgAdapter
from FlopyAdapter.MtPackages.SsmAdapter import SsmAdapter


class InowasFlopyImportAdapter:
    """
    The Flopy Import Class.
    Serialize given Modflow and MT3D models to json format and
    writes it to specified .json file
    """

    def __init__(self, model_ws, json_file, mf_namfile=None, mt_namfile=None,
                 type_="flopy_calculation", version="3.2.6", calculation_id="default",
                 author="default", project="default", model_id="default",
                 run_model=True, write_input=True):

        self.model_data = {
            "author": author,
            "project": project,
            "type": type_,
            "version": version,
            "calculation_id": calculation_id,
            "model_id": model_id,
            "write_input": write_input,
            "run_model": run_model,
            "data": {}
        }
        self._report = ''
        self.json_file = json_file
        self.mf_packages = []
        self.mt_packages = []
        self.mf_model, self.mt_model = None, None

        if mf_namfile is not None:
            self.mf_model = flopy.modflow.Modflow.load(os.path.join(model_ws, mf_namfile))
            self.model_data["data"]["mf"] = {"packages": ["MF"]}
            self.model_data["data"]["mf"]["packages"] += self.mf_model.get_package_list()
            self.model_data["data"]["mf"]["write_input"] = write_input
            self.model_data["data"]["mf"]["run_model"] = run_model
        if mt_namfile is not None:
            self.mt_model = flopy.mt3d.Mt3dms.load(os.path.join(model_ws, mt_namfile))
            self.model_data["data"]["mt"] = {"packages": ["MT"]}
            self.model_data["data"]["mt"]["packages"] += self.mt_model.get_package_list()
            self.model_data["data"]["mt"]["write_input"] = write_input
            self.model_data["data"]["mt"]["run_model"] = run_model

    @staticmethod
    def np_type_translate(obj):
        try:
            return obj.item()
        except:
            raise TypeError('Object %s is not JSON serializable and not Numpy dtype' % type(obj))

    def serialize(self):
        # Encode packages content to JSON
        for package_name in self.model_data["data"]["mf"]["packages"] + self.model_data["data"]["mt"]["packages"]:
            try:
                self.read_packages(name=package_name, data=self.model_data["data"])
                self._report += "Successfully read the package: %s \n" % package_name
            except:
                self._report += "Could not read package: %s \n" % package_name

        try:
            with open(self.json_file, 'w') as f:
                json.dump(self.model_data, f, default=self.np_type_translate)
            self._report += "Model input saved to %s \n" % self.json_file
        except:
            self._report += "Could not save input to %s \n" % self.json_file

    def read_packages(self, name, data):
        # Modflow packages
        if name == 'MF':
            data["mf"][name] = MfAdapter(data=None).read_package(self.mf_model)
        if name == 'DIS':
            data["mf"][name] = DisAdapter(data=None).read_package(self.mf_model.get_package(name))
        if name == 'BAS' or name == 'BAS6':
            data["mf"][name] = BasAdapter(data=None).read_package(self.mf_model.get_package(name))
        if name == 'LPF':
            data["mf"][name] = LpfAdapter(data=None).read_package(self.mf_model.get_package(name))
        if name == 'PCG':
            data["mf"][name] = PcgAdapter(data=None).read_package(self.mf_model.get_package(name))
        if name == 'OC':
            data["mf"][name] = {}
            # data["mf"][name] = OcAdapter(data=None).read_package(self.mf_model.get_package(name))
        if name == 'WEL':
            data["mf"][name] = WelAdapter(data=None).read_package(self.mf_model.get_package(name))
        if name == 'CHD':
            data["mf"][name] = ChdAdapter(data=None).read_package(self.mf_model.get_package(name))
        if name == 'LMT' or name == 'LMT6':
            data["mf"][name] = LmtAdapter(data=None).read_package(self.mf_model.get_package(name))
        if name == 'NWT':
            data["mf"][name] = NwtAdapter(data=None).read_package(self.mf_model.get_package(name))
        # Not checked:
        # if name == 'UPW':
        #     data["mf"][name] = UpwAdapter(data=None).read_package(self.mf_model.get_package(name))
        # if name == 'RIV':
        #     data["mf"][name] = RivAdapter(data=None).read_package(self.mf_model.get_package(name))
        # if name == 'RCH':
        #     data["mf"][name] = RchAdapter(data=None).read_package(self.mf_model.get_package(name))
        # if name == 'GHB':
        #     data["mf"][name] = GhbAdapter(data=None).read_package(self.mf_model.get_package(name))

        # MT3D packages
        if name == 'MT':
            data["mt"][name] = MtAdapter(data=None).read_package(self.mt_model)
        if name == 'ADV':
            data["mt"][name] = AdvAdapter(data=None).read_package(self.mt_model.get_package(name))
        if name == 'BTN':
            data["mt"][name] = BtnAdapter(data=None).read_package(self.mt_model.get_package(name))
        if name == 'DSP':
            data["mt"][name] = DspAdapter(data=None).read_package(self.mt_model.get_package(name))
        if name == 'GCG':
            data["mt"][name] = GcgAdapter(data=None).read_package(self.mt_model.get_package(name))
        if name == 'SSM':
            data["mt"][name] = SsmAdapter(data=None).read_package(self.mt_model.get_package(name))
        # Not checked:
        # if name == 'LKT':
        #     data["mt"][name] = LktAdapter(data=None).read_package(self.mt_model.get_package(name))
        # if name == 'PHC':
        #     data["mt"][name] = PhcAdapter(data=None).read_package(self.mt_model.get_package(name))
        # if name == 'RCT':
        #     data["mt"][name] = RctAdapter(data=None).read_package(self.mt_model.get_package(name))
        # if name == 'SFT':
        #     data["mt"][name] = SftAdapter(data=None).read_package(self.mt_model.get_package(name))
        # if name == 'TOB':
        #     data["mt"][name] = TobAdapter(data=None).read_package(self.mt_model.get_package(name))
        # if name == 'UZT':
        #     data["mt"][name] = UztAdapter(data=None).read_package(self.mt_model.get_package(name))

    @property
    def response_message(self):
        return self._report



"""
This module is an intermediate layer between flopy version 3.2
and the inowas-modflow-configuration format.

Author: Ralf Junghanns
EMail: ralf.junghanns@gmail.com
"""

from FlopyAdapter.Read.ReadBudget import ReadBudget
from FlopyAdapter.Read.ReadConcentration import ReadConcentration
from FlopyAdapter.Read.ReadDrawdown import ReadDrawdown
from FlopyAdapter.Read.ReadHead import ReadHead
from FlopyAdapter.Read.ReadFile import ReadFile


class InowasFlopyReadAdapter:
    """The Flopy Class"""

    _request = None
    _projectfolder = None
    _version = None

    def __init__(self, version, projectfolder, request):
        self._request = request
        self._projectfolder = projectfolder
        self._version = version
        pass

    def read_head(self, totim, layer):
        head_file = ReadHead(self._projectfolder)
        return head_file.read_layer(totim=totim, layer=layer)

    def read_head_ts(self, layer, row, column):
        head_file = ReadHead(self._projectfolder)
        return head_file.read_ts(layer=layer, row=row, column=column)

    def read_concentration(self, totim, layer):
        concentration_file = ReadConcentration(self._projectfolder)
        return concentration_file.read_layer(totim=totim, layer=layer)

    def read_concentration_ts(self, layer, row, column):
        concentration_file = ReadConcentration(self._projectfolder)
        return concentration_file.read_ts(layer=layer, row=row, column=column)

    def read_drawdown(self, totim, layer):
        drawdown_file = ReadDrawdown(self._projectfolder)
        return drawdown_file.read_layer(totim=totim, layer=layer)

    def read_drawdown_ts(self, layer, row, column):
        drawdown_file = ReadDrawdown(self._projectfolder)
        return drawdown_file.read_ts(layer=layer, row=row, column=column)

    def read_cumulative_budget(self, totim):
        budget_file = ReadBudget(self._projectfolder)
        return budget_file.read_cumulative_budget(totim=totim)

    def read_incremental_budget(self, totim):
        budget_file = ReadBudget(self._projectfolder)
        return budget_file.read_incremental_budget(totim=totim)

    def read_file(self, extension):
        namfile = ReadFile(self._projectfolder)
        return namfile.read_file(extension)

    def read_file_list(self):
        namfile = ReadFile(self._projectfolder)
        return namfile.read_file_list()

    def response(self):

        data = None
        request = self._request

        if 'budget' in request:
            if request['budget']['type'] == 'cumulative':
                totim = request['budget']['totim']
                data = self.read_cumulative_budget(totim=totim)

            if request['budget']['type'] == 'incremental':
                totim = self._request['totim']
                data = self.read_incremental_budget(totim=totim)

        if 'layerdata' in request:
            if request['layerdata']['type'] == 'concentration':
                totim = request['layerdata']['totim']
                layer = request['layerdata']['layer']
                data = self.read_concentration(totim=totim, layer=layer)

            if request['layerdata']['type'] == 'drawdown':
                totim = request['layerdata']['totim']
                layer = request['layerdata']['layer']
                data = self.read_drawdown(totim=totim, layer=layer)

            if request['layerdata']['type'] == 'head':
                totim = request['layerdata']['totim']
                layer = request['layerdata']['layer']
                data = self.read_head(totim=totim, layer=layer)

        if 'file' in request:
            data = [self.read_file(request['file'])]

        if 'filelist' in request:
            data = self.read_file_list()

        if 'timeseries' in request:
            if request['timeseries']['type'] == 'concentration':
                layer = request['timeseries']['layer']
                row = request['timeseries']['row']
                column = request['timeseries']['column']
                data = self.read_concentration_ts(layer=layer, row=row, column=column)

            if request['timeseries']['type'] == 'drawdown':
                layer = request['timeseries']['layer']
                row = request['timeseries']['row']
                column = request['timeseries']['column']
                data = self.read_drawdown_ts(layer=layer, row=row, column=column)

            if request['timeseries']['type'] == 'head':
                layer = request['timeseries']['layer']
                row = request['timeseries']['row']
                column = request['timeseries']['column']
                data = self.read_head_ts(layer=layer, row=row, column=column)

        if data is not None:
            return dict(
                status_code=200,
                request=request,
                response=data
            )

        return dict(
            status_code=500,
            message="Internal Server Error. Request data does not fit."
        )



import flopy.seawat as seawat


class SwtAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self):
        content = self.merge()
        return seawat.Seawat(
            **content
        )

    @staticmethod
    def default():
        return {
            "modelname": 'swttest',
            "namefile_ext": 'nam',
            "modflowmodel": None,
            "mt3dmodel": None,
            "version": 'seawat',
            "exe_name": 'swtv4',
            "structured": True,
            "listunit": 2,
            "model_ws": '.',
            "external_path": None,
            "verbose": False,
            "load": True,
            "silent": 0
        }

    @staticmethod
    def read_package(package):
        return {
            "modelname": package.modelname,
            "namefile_ext": package.namefile,
            "modflowmodel": package.modflowmodel,
            "mt3dmodel": package.mt3dmodel,
            "version": package.version,
            "exe_name": package.exe,
            "structured": package.structured,
            "listunit": package.listunit,
            "model_ws": package.model,
            "external_path": package.external,
            "verbose": package.verbose,
            "load": package.load,
            "silent": package.silent
        }



import flopy.seawat as seawat


class VdfAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _swt):
        content = self.merge()
        return seawat.SeawatVdf(
            _swt,
            **content
        )

    @staticmethod
    def default():
        default = {
            "mtdnconc": 1,
            "mfnadvfd": 1,
            "nswtcpl": 1,
            "iwtable": 1,
            "densemin": 1.0,
            "densemax": 1.025,
            "dnscrit": 0.01,
            "denseref": 1.0,
            "denseslp": 0.025,
            "drhodc": 0.01,
            "crhoref": 0,
            "firstdt": 0.001,
            "indense": 1,
            "dense": 1.0,
            "nsrhoeos": 1,
            "drhodprhd": 0.00446,
            "prhdref": 0.0,
            "extension": 'vdf',
            "unitnumber": None
        }
        return default

    @staticmethod
    def read_package(package):
        return {
            "mtdnconc": package.mtdnconc,
            "mfnadvfd": package.mfnadvfd,
            "nswtcpl": package.nswtcpl,
            "iwtable": package.iwtable,
            "densemin": package.densemin,
            "densemax": package.densemax,
            "dnscrit": package.dnscrit,
            "denseref": package.denseref,
            "denseslp": package.denseslp,
            "crhoref": package.crhoref,
            "firstdt": package.firstdt,
            "indense": package.indense,
            "dense": package.dense,
            "nsrhoeos": package.nsrhoeos,
            "drhodprhd": package.drhodprhd,
            "prhdref": package.prhdref,
            "extension": package.extension,
            "unitnumber": package.unitnumber,
            "filenames": package.filenames
        }



import os


class ReadFile:
    _filename = None
    _workspace = None

    def __init__(self, workspace):
        self._workspace = workspace
        pass

    def read_file(self, extension):

        workspace = self._workspace
        for file in os.listdir(self._workspace):
            if file.endswith("." + extension):
                self._filename = os.path.join(self._workspace, file)

        try:
            with open(self._filename) as f:
                return f.read()
        except:
            return ("Error reading file with extension '" + str(extension)) + "' in workspace '" + str(workspace) + "'."

    def read_file_list(self):
        workspace = self._workspace
        try:
            return os.listdir(workspace)
        except:
            return "Error reading files of workspace '" + str(workspace) + "'."



import flopy.modpath as mp


class MpAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self):
        content = self.merge()
        return mp.Modpath(
            **content
        )

    @staticmethod
    def default():
        return {
            "modelname": 'modpathtest',
            "simfile_ext": 'mpsim',
            "namefile_ext": 'mpnam',
            "version": 'modpath',
            "exe_name": 'mp6.exe',
            "modflowmodel": None,
            "dis_file": None,
            "dis_unit": 87,
            "head_file": None,
            "budget_file": None,
            "model_ws": None,
            "external_path": None,
            "verbose": False,
            "load": True,
            "listunit": 7
        }

    @staticmethod
    def read_package(package):
        return {
            "modelname": package.modelname,
            "simfile_ext": package.simfile_ext,
            "namefile_ext": package.namefile_ext,
            "version": package.version,
            "exe_name": package.exe_name,
            "modflowmodel": package.modflowmodel,
            "dis_file": package.dis_file,
            "dis_unit": package.dis_unit,
            "head_file": package.head_file,
            "budget_file": package.budget_file,
            "model_ws": package.model_ws,
            "external_path": package.external_path,
            "verbose": package.verbose,
            "load": package.load,
            "listunit": package.listunit
        }



import flopy.mt3d as mt


class SftAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            if key == 'sf_stress_period_data':
                default[key] = self.to_dict(self._data[key])
                continue

            default[key] = self._data[key]
        return default

    def to_dict(self, data):
        if type(data) == list:
            spd_dict = {}
            for stress_period, record in enumerate(data):
                spd_dict[stress_period] = record
            return spd_dict
        return data

    def get_package(self, _mt):
        content = self.merge()
        return mt.Mt3dSft(
            _mt,
            **content
        )

    @staticmethod
    def default():
        default = {
            "nsfinit": 0,
            "mxsfbc": 0,
            "icbcsf": 0,
            "ioutobs": None,
            "ietsfr": 0,
            "isfsolv": 1,
            "wimp": 0.5,
            "wups": 1.0,
            "cclosesf": 1e-06,
            "mxitersf": 10,
            "crntsf": 1.0,
            "iprtxmd": 0,
            "coldsf": 0.0,
            "dispsf": 0.0,
            "nobssf": 0,
            "obs_sf": None,
            "sf_stress_period_data": None,
            "unitnumber": None,
            "filenames": None,
            "dtype": None,
            "extension": 'sft'
        }
        return default

    @staticmethod
    def read_package(package):
        content = {
            "nsfinit": package.nsfinit,
            "mxsfbc": package.mxsfbc,
            "icbcsf": package.icbcsf,
            "ioutobs": package.ioutobs,
            "ietsfr": package.ietsfr,
            "isfsolv": package.isfsolv,
            "wimp": package.wimp,
            "wups": package.wups,
            "cclosesf": package.cclosesf,
            "mxitersf": package.mxitersf,
            "crntsf": package.crntsf,
            "iprtxmd": package.iprtxmd,
            "coldsf": package.coldsf,
            "dispsf": package.dispsf,
            "nobssf": package.nobssf,
            "obs_sf": package.obs_sf,
            "sf_stress_period_data": package.sf_stress_period_data,
            "unitnumber": package.unitnumber,
            "filenames": package.filenames,
            "dtype": package.dtype,
            "extension": package.extension
        }
        return content



import flopy.mt3d as mt


class PhcAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mt):
        content = self.merge()
        return mt.Mt3dPhc(
            _mt,
            **content
        )

    @staticmethod
    def default():
        return {
            "os": 2,
            "temp": 25,
            "asbin": 0,
            "eps_aqu": 0,
            "eps_ph": 0,
            "scr_output": 1,
            "cb_offset": 0,
            "smse": ['pH', 'pe'],
            "mine": [],
            "ie": [],
            "surf": [],
            "mobkin": [],
            "minkin": [],
            "surfkin": [],
            "imobkin": [],
            "extension": 'phc',
            "unitnumber": None
        }

    @staticmethod
    def read_package(package):
        return {
            "os": package.os,
            "temp": package.temp,
            "asbin": package.asbin,
            "eps_aqu": package.eps_aqu,
            "eps_ph": package.eps_ph,
            "scr_output": package.scr_output,
            "cb_offset": package.cb_offset,
            "smse": package.smse,
            "mine": package.mine,
            "ie": package.ie,
            "surf": package.surf,
            "mobkin": package.mobkin,
            "minkin": package.minkin,
            "surfkin": package.surfkin,
            "imobkin": package.imobkin,
            "extension": package.extension,
            "unitnumber": package.unitnumber
        }



import flopy.modflow as mf


class HobAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mf):
        content = self.merge()
        # noinspection PyTypeChecker
        content["obs_data"] = self.map_obs_data(_mf, content["obs_data"])
        return mf.ModflowHob(
            _mf,
            **content
        )

    @staticmethod
    def map_obs_data(model, observations):
        obs = []

        for o in observations:
            print(o)
            obs.append(mf.HeadObservation(
                model,
                layer=o['layer'],
                row=o['row'],
                column=o['column'],
                time_series_data=o['time_series_data']
            ))

        return obs

    @staticmethod
    def default():
        default = {
            "iuhobsv": 1051,
            "hobdry": 0,
            "tomulth": 1.0,
            "obs_data": None,
            "hobname": None,
            "extension": 'hob',
            "unitnumber": None,
            "filenames": None
        }
        return default

    @staticmethod
    def read_package(package):
        content = {
            "iuhobsv": package.iuhobsv,
            "hobdry": package.hobdry,
            "tomulth": package.tomulth,
            "obs_data": package.obs_data,
            "hobname": package.hobname,
            "extension": package.extension[0],
            "unitnumber": package.unit_number[0],
            "filenames": package.filenames
        }
        return content



import flopy.modflow as mf


class MfAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self):
        content = self.merge()
        return mf.Modflow(
            **content
        )

    @staticmethod
    def default():
        default = {
            "modelname": "modflowtest",
            "namefile_ext": 'nam',
            "version": "mf2005",
            "exe_name": "mf2005.exe",
            "structured": True,
            "listunit": 2,
            "model_ws": '.',
            "external_path": None,
            "verbose": False
        }
        return default

    @staticmethod
    def read_package(package):
        content = {
            # "modelname": package.modelname,#None
            "namefile_ext": package.namefile_ext,
            "version": package.version,
            "exe_name": package.exe_name,
            "structured": package.structured,
            # "listunit": package.listunit,
            "model_ws": package.model_ws,
            "external_path": package.external_path,
            "verbose": package.verbose
        }
        return content



import os
import flopy.utils.binaryfile as bf


class ReadHead:
    _filename = None

    def __init__(self, workspace):
        for file in os.listdir(workspace):
            if file.endswith(".hds"):
                self._filename = os.path.join(workspace, file)
        pass

    def read_times(self):
        try:
            heads = bf.HeadFile(filename=self._filename, precision='single')
            return heads.get_times()
        except:
            return []

    def read_number_of_layers(self):
        try:
            heads = bf.HeadFile(filename=self._filename, precision='single')
            number_of_layers = heads.get_data().shape[0]
            return number_of_layers
        except:
            return 0

    def read_layer(self, totim, layer):
        try:
            heads = bf.HeadFile(filename=self._filename, precision='single')
            data = heads.get_data(totim=totim, mflay=layer).tolist()
            for i in range(len(data)):
                for j in range(len(data[i])):
                    data[i][j] = round(data[i][j], 2)
                    if data[i][j] < -999:
                        data[i][j] = None
            return data
        except:
            return []

    def read_ts(self, layer, row, column):
        try:
            heads = bf.HeadFile(filename=self._filename, precision='single')
            return heads.get_ts(idx=(layer, row, column)).tolist()
        except:
            return []



import flopy.modpath as mp


class BasAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            if not key.startswith('_'):
                default[key] = self._data[key]
        return default

    def get_package(self, _mp):
        content = self.merge()
        return mp.ModpathBas(
            _mp,
            **content
        )

    @staticmethod
    def default():
        return {
            "hnoflo": -9999.0,
            "hdry": -8888.0,
            "def_face_ct": 0,
            "bud_label": None,
            "def_iface": None,
            "laytyp": 0,
            "ibound": 1,
            "prsity": 0.3,
            "prsityCB": 0.3,
            "extension": 'mpbas',
            "unitnumber": 86
        }

    @staticmethod
    def read_package(package):
        return {
            "hnoflo": package.hnoflo,
            "hdry": package.hdry,
            "def_face_ct": package.def_face_ct,
            "bud_label": package.bud_label,
            "def_iface": package.def_iface,
            "laytyp": package.laytyp,
            "ibound": package.ibound,
            "prsity": package.prsity,
            "prsityCB": package.prsityCB,
            "extension": package.extension,
            "unitnumber": package.unitnumber,
        }



import flopy.seawat as seawat


class VscAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _swt):
        content = self.merge()
        return seawat.SeawatVsc(
            _swt,
            **content
        )

    @staticmethod
    def default():
        return {
            "mt3dmuflg": -1,
            "viscmin": 0.0,
            "viscmax": 0.0,
            "viscref": 0.0008904,
            "nsmueos": 0,
            "mutempopt": 2,
            "mtmuspec": 1,
            "dmudc": 1.923e-06,
            "cmuref": 0.0,
            "mtmutempspec": 1,
            "amucoeff": None,
            "invisc": -1,
            "visc": -1,
            "extension": 'vsc',
            "unitnumber": None
        }

    @staticmethod
    def read_package(package):
        return {
            "mt3dmuflg": package.mt3dmuflg,
            "viscmin": package.viscmin,
            "viscmax": package.viscmax,
            "viscref": package.viscref,
            "nsmueos": package.nsmueos,
            "mutempopt": package.mutempopt,
            "mtmuspec": package.mtmuspec,
            "dmudc": package.dmudc,
            "cmuref": package.cmuref,
            "mtmutempspec": package.mtmutempspec,
            "amucoeff": package.amucoeff,
            "invisc": package.invisc,
            "visc": package.visc,
            "extension": package.extension,
            "unitnumber": package.unitnumber,
            "filenames": package.filenames
        }



import flopy.modflow as mf


class OcAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]

        if 'stress_period_data' in self._data:
            default['stress_period_data'] = self.get_stress_period_data(self._data['stress_period_data'])

        return default

    def get_package(self, _mf):
        content = self.merge()
        return mf.ModflowOc(
            _mf,
            **content
        )

    @staticmethod
    def default():
        default = {
            "ihedfm": 0,
            "iddnfm": 0,
            "chedfm": None,
            "cddnfm": None,
            "cboufm": None,
            "compact": True,
            "stress_period_data": None,
            "extension": ['oc', 'hds', 'ddn', 'cbc'],
            "unitnumber": [14, 51, 52, 53]
        }

        return default

    @staticmethod
    def get_stress_period_data(stress_periods):
        if stress_periods is None:
            return stress_periods

        stress_period_data = {}
        for stress_period in stress_periods:
            sp = stress_period[0][0]
            ts = stress_period[0][1]
            oc_data = stress_period[1]
            stress_period_data[(sp, ts)] = oc_data

        return stress_period_data

    @staticmethod
    def read_package(package):
        content = {
            "ihedfm": package.ihedfm,
            "iddnfm": package.iddnfm,
            "chedfm": package.chedfm,
            "cddnfm": package.cddnfm,  # None
            "cboufm": package.cboufm,  # None
            "compact": package.compact,
            # stress period data dict keys transformed from tuple to string to be json serializable
            "stress_period_data": {str(k): v for k, v in package.stress_period_data.items()},
            "extension": package.extension[0],
            "unitnumber": package.unit_number[0]
        }
        return content



from .InowasFlopyReadAdapter import InowasFlopyReadAdapter
from .InowasFlopyReadFitness import InowasFlopyReadFitness
from .InowasModflowReadAdapter import InowasModflowReadAdapter
from .ReadBudget import ReadBudget
from .ReadConcentration import ReadConcentration
from .ReadDrawdown import ReadDrawdown
from .ReadFile import ReadFile
from .ReadHead import ReadHead



import flopy.modflow as mf


class EvtAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mf):
        content = self.merge()
        return mf.ModflowEvt(
            _mf,
            **content
        )

    @staticmethod
    def default():
        return {
            "nevtop": 3,
            "ipakcb": None,
            "surf": 0.,
            "evtr": 1e-3,
            "exdp": 1.,
            "ievt": 1,
            "extension": 'evt',
            "unitnumber": None,
            "filenames": None,
            "external": True
        }

    @staticmethod
    def read_package(package):
        return {
            "nevtop": package.nevtop,
            "ipakcb": package.ipakcb,
            "surf": package.surf,
            "evtr": package.evtr,
            "exdp": package.exdp,
            "ievt": package.ievt,
            "extension": package.extension,
            "unitnumber": package.unitnumber,
            "filenames": package.filenames,
            "external": package.external
        }



"""
Calculation of objective values of a model 

Author: Aybulat Fatkhutdinov
"""

import os
import math
import numpy as np
import flopy


class InowasFlopyReadFitness:
    """Calculation of objective values of a model """

    def __init__(self, optimization_data, flopy_adapter):

        self.optimization_data = optimization_data

        self.dis_package = flopy_adapter._mf.get_package('DIS')
        self.model_ws = flopy_adapter._mf.model_ws
        self.model_name = flopy_adapter._mf.namefile.split('.')[0]
        self.objects = self.optimization_data['objects']

        objectives_values = self.read_objectives()
        constraints_exceeded = self.check_constraints()

        if True in constraints_exceeded or None in objectives_values:
            self.fitness = [obj["penalty_value"] for obj in optimization_data["objectives"]]
        else:
            self.fitness = objectives_values


    def get_fitness(self):

        return self.fitness

    def read_objectives(self):
        "Returnes fitnes list"
        fitness = []

        for objective in self.optimization_data["objectives"]:

            if objective["type"] == "concentration":
                mask = self.make_mask(
                    objective["location"], self.objects, self.dis_package
                )
                value = self.read_concentration(objective, mask, self.model_ws, self.model_name)

            elif objective["type"] == "head":
                mask = self.make_mask(
                    objective["location"], self.objects, self.dis_package
                )
                value = self.read_head(objective, mask, self.model_ws, self.model_name)
          

            elif objective["type"] == "flux":
                value = self.read_flux(objective, self.objects)

            
            elif objective["type"] == "input_concentration":
                value = self.read_input_concentration(objective, self.objects)
            
            value = self.summary(value, objective["summary_method"])
            fitness.append(value.item())

        return fitness
    
    def check_constraints(self):
        """Returns a list of penalty values"""
        constraints_exceeded = []

        for constraint in self.optimization_data["constraints"]:

            if constraint["type"] == 'head':
                mask = self.make_mask(
                    constraint["location"], self.objects, self.dis_package    
                )
                value = self.read_head(
                    constraint, mask, self.model_ws, self.model_name
                )
   
            elif constraint["type"] == 'concentration':
                mask = self.make_mask(
                    constraint["location"], self.objects, self.dis_package    
                )
                value = self.read_concentration(
                    constraint, mask, self.model_ws, self.model_name
                )
            
            elif constraint["type"] == "flux":
                value = self.read_flux(
                    constraint, self.objects
                )
            
            elif constraint["type"] == "input_concentrations":
                value = self.read_input_concentration(
                    constraint, self.objects
                )
            
            value = self.summary(value, constraint["summary_method"])
            
            if constraint["operator"] == "less":
                if value > constraint["value"]:
                    print("Constraint value {} exceeded max value {}, penalty will be assigned".format(value, constraint["value"]))
                    constraints_exceeded.append(True)
                else:
                    constraints_exceeded.append(False)
                
            elif constraint["operator"] == "more":
                if value < constraint["value"]:
                    print("Constraint value {} lower than min value {}, penalty will be assigned".format(value, constraint["value"]))
                    constraints_exceeded.append(True)
                else:
                    constraints_exceeded.append(False)

        return constraints_exceeded
    
    @staticmethod
    def summary(result, method):
        if method == 'mean':
            result = np.nanmean(result)
        elif method == 'max':
            result = np.max(result)
        elif method == 'min':
            result = np.min(result)
        else:
            print("Unknown summary method {}. Using max".format(method))
            result = np.max(result)
        
        return result


    @staticmethod
    def read_head(data, mask, model_ws, model_name):
        "Reads head file"

        print('Read head values at location: {}'.format(data['location']))
        
        try:
            head_file_object = flopy.utils.HeadFile(
                os.path.join(model_ws, model_name) + '.hds')
            head = head_file_object.get_alldata(
                nodata=-9999
                )
            head = head[mask]

            head_file_object.close()

        except:
            print('Head file of the model: '+model_name+' could not be opened')

        return head
    
    @staticmethod
    def read_concentration(data, mask, model_ws, model_name):
        "Reads concentrations file"

        print('Read concentration values at location: {}'.format(data['location']))

        try:
            conc_file_object = flopy.utils.UcnFile(
                os.path.join(model_ws, data["conc_file_name"]))
            conc = conc_file_object.get_alldata(
                nodata=-9999
                )
            conc = conc[mask]

            conc_file_object.close()
        
        except:
            print('Concentrations file of the model: '+model_name+' could not be opened')
            return None

        return conc
    
    @staticmethod
    def read_flux(data, objects):
        "Reads wel fluxes"

        print('Read flux values at location: {}'.format(data['location']))

        fluxes = np.array([])

        try:
            obj_ids = data["location"]["objects"]
        except KeyError:
            print("ERROR! Objective location of type Flux has to be an Object!")
            return None

        for obj in objects:
            if obj['id'] in obj_ids:
                obj_fluxes = []
                for period_data in obj['flux'].values():
                    obj_fluxes.append(period_data['result'])
        
                fluxes = np.hstack(
                    (fluxes,
                    np.array(obj_fluxes))
                )

        return fluxes
    
    @staticmethod
    def read_input_concentration(data, objects):

        print('Read input_concentration values at location: {}'.format(data['location']))

        input_concentrations = np.array([])

        try:
            component = data["component"]
        except KeyError:
            print("ERROR! Concentration component for the Objective of type input_concentrations is not defined!")
            return None
        
        try:
            obj_ids = data["location"]["objects"]
        except KeyError:
            print("ERROR! Objective location of type input_concentrations has to be an Object!")
            return None
        
        for obj in objects:
            if obj['id'] in obj_ids:
                obj_concentrations = []
                for period_data in obj['concentration'].values():
                    obj_concentrations.append(period_data[component]['result'])
                input_concentrations = np.hstack(
                    (input_concentrations, 
                    np.array(obj_concentrations))
                )
        

        return input_concentrations
    
    @staticmethod
    def read_distance(data, objects):
        """Returns distance between two groups of objects"""

        print('Read distance between {} and {}'.format(data['location_1'], data['location_2']))

        location_1 = data["location_1"]
        location_2 = data["location_2"]
        
        objects_1 = None
        objects_2 = None

        if location_1['type'] == 'object':
            objects_1 = [
                obj for id_, obj in objects.items() if id_ in location_1['objects_ids']
            ]

        if location_2['type'] == 'object':
            objects_2 =[
                obj for id_, obj in objects.items() if id_ in location_2['objects_ids']
            ]
        
        distances = []
        if objects_1 is not None:
            for obj_1 in objects_1:
                if objects_2 is not None:
                    for obj_2 in objects_2:
                        dx = float(abs(obj_2['position']['col']['result'] - obj_1['position']['col']['result']))
                        dy = float(abs(obj_2['position']['row']['result'] - obj_1['position']['row']['result']))
                        dz = float(abs(obj_2['position']['lay']['result'] - obj_1['position']['lay']['result']))
                        distances.append(math.sqrt((dx**2) + (dy**2) + (dz**2)))
                else:
                    dx = float(abs(location_2['lay_row_col'][2] - obj_1['position']['col']['result']))
                    dy = float(abs(location_2['lay_row_col'][1] - obj_1['position']['row']['result']))
                    dz = float(abs(location_2['lay_row_col'][0] - obj_1['position']['lay']['result']))
                    distances.append(math.sqrt((dx**2) + (dy**2) + (dz**2)))
        else:
            if objects_2 is not None:
                for obj_2 in objects_2:
                    dx = float(abs(obj_2['position']['col']['result'] - location_1['lay_row_col'][2]))
                    dy = float(abs(obj_2['position']['row']['result'] - location_1['lay_row_col'][1]))
                    dz = float(abs(obj_2['position']['lay']['result'] - location_1['lay_row_col'][0]))
                    distances.append(math.sqrt((dx**2) + (dy**2) + (dz**2)))
            else:
                dx = float(abs(location_2['lay_row_col'][2]-location_1['lay_row_col'][2]))
                dy = float(abs(location_2['lay_row_col'][1]-location_1['lay_row_col'][1]))
                dz = float(abs(location_2['lay_row_col'][0]-location_1['lay_row_col'][0]))
                distances.append(math.sqrt((dx**2) + (dy**2) + (dz**2)))

        distances = np.array(distances)
        
        return distances

    @staticmethod
    def make_mask(location, objects, dis_package):
        "Returns an array mask of location that has nper,nlay,nrow,ncol dimensions"

        print('Making mask array for location: {}'.format(location))
        nstp_flat = dis_package.nstp.array.sum()
        nrow = dis_package.nrow
        ncol = dis_package.ncol
        nlay = dis_package.nlay

        if location["type"] == 'bbox':
            try:
                per_min = location['per_min']
            except KeyError:
                per_min = 0

            try:
                per_max = location['per_max']
            except KeyError:
                per_max = nstp_flat

            try:
                lay_min = location['lay_min']
            except KeyError:
                lay_min = 0

            try:
                lay_max = location['lay_max']
            except KeyError:
                lay_max = nlay

            try:
                col_min = location['col_min']
            except KeyError:
                col_min = 0

            try:
                col_max = location['col_max']
            except KeyError:
                col_max = ncol

            try:
                row_min = location['row_min']
            except KeyError:
                row_min = 0

            try:
                row_max = location['row_max']
            except KeyError:
                row_max = nrow

            if per_min == per_max:
                per_max += 1
            if lay_min == lay_max:
                lay_max += 1
            if row_min == row_max:
                row_max += 1
            if col_min == col_max:
                col_max += 1
        
            mask = np.zeros((nstp_flat, nlay, nrow, ncol), dtype=bool)
            mask[
                per_min:per_max,
                lay_min:lay_max,
                row_min:row_max,
                col_min:col_max
            ] = True

        elif location["type"] == 'object':
            lays = []
            rows = []
            cols = []
            for obj in objects:
                if obj['id'] in location['objects']:
                    lays.append(obj['position']['lay']['result'])
                    rows.append(obj['position']['row']['result'])
                    cols.append(obj['position']['col']['result'])

            mask = np.zeros((nstp_flat, nlay, nrow, ncol), dtype=bool)
            mask[:,lays,rows,cols] = True
        
        return mask


import flopy.mt3d as mt


class RctAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mt):
        content = self.merge()
        return mt.Mt3dRct(
            _mt,
            **content
        )

    @staticmethod
    def default():
        default = {
            "isothm": 0,
            "ireact": 0,
            "igetsc": 1,
            "rhob": None,
            "prsity2": None,
            "srconc": None,
            "sp1": None,
            "sp2": None,
            "rc1": None,
            "rc2": None,
            "extension": 'rct',
            "unitnumber": None,
            "filenames": None
        }
        return default

    @staticmethod
    def read_package(package):
        content = {
            "isothm": package.isothm,
            "ireact": package.ireact,
            "igetsc": package.igetsc,
            "rhob": package.rhob,
            "prsity2": package.prsity2,
            "srconc": package.srconc,
            "sp1": package.sp1,
            "sp2": package.sp2,
            "rc1": package.rc1,
            "rc2": package.rc2,
            "extension": package.extension,
            "unitnumber": package.unitnumber,
            "filenames": package.filenames
        }
        return content



import os
from flopy.utils.mflistfile import MfListBudget


class ReadBudget:
    _filename = None

    def __init__(self, workspace):
        for file in os.listdir(workspace):
            if file.endswith(".list"):
                self._filename = os.path.join(workspace, file)
        pass

    def read_times(self):
        try:
            mf_list = MfListBudget(self._filename)
            if mf_list.get_times():
                return mf_list.get_times()
            else:
                return []
        except:
            return []

    def read_budget_by_totim(self, totim=0, incremental=False):
        try:
            mf_list = MfListBudget(self._filename)
            budget = mf_list.get_data(totim=totim, incremental=incremental)
            values = {}
            for x in budget:
                param = str(x[2].decode('UTF-8'))
                values[param] = float(str(x[1]))
            return values
        except:
            return []

    def read_budget_by_idx(self, idx=0, incremental=False):
        try:
            mf_list = MfListBudget(self._filename)
            budget = mf_list.get_data(idx=idx, incremental=incremental)
            values = {}
            for x in budget:
                param = str(x[2].decode('UTF-8'))
                values[param] = float(str(x[1]))
            return values
        except:
            return []

    def read_budget_by_kstpkper(self, kstpkper=(0, 0), incremental=False):
        try:
            mf_list = MfListBudget(self._filename)
            budget = mf_list.get_data(kstpkper=kstpkper, incremental=incremental)
            values = {}
            for x in budget:
                param = str(x[2].decode('UTF-8'))
                values[param] = float(str(x[1]))
            return values
        except:
            return []



import flopy as fp
import geojson
import glob
import numpy as np
import os
import rasterio.features
import sys
import utm
from flopy.discretization import StructuredGrid
from pyproj import Transformer
import FlopyAdapter.MfPackages as fpa


class InvalidArgumentException(Exception):
    pass


class InvalidModelUnitException(Exception):
    pass


class InowasModflowReadAdapter:
    _ws = None
    _mf = None

    @staticmethod
    def load(path):

        abspath = os.path.abspath(os.path.join(path))
        if not os.path.exists(abspath):
            raise FileNotFoundError('Path not found: ' + path)

        if len(glob.glob1(abspath, "*.nam")) == 0 and len(glob.glob1(abspath, "*.mfn")) == 0:
            raise FileNotFoundError('Modflow name file with ending .nam or .mfn not found')

        orig_stdout = sys.stdout
        f = open(os.path.join(abspath, 'load.log'), 'w')
        sys.stdout = f

        instance = InowasModflowReadAdapter()
        instance._ws = path

        name_file = ''
        if len(glob.glob1(abspath, "*.nam")) > 0:
            name_file = glob.glob1(abspath, "*.nam")[0]

        if len(glob.glob1(abspath, "*.mfn")) > 0:
            name_file = glob.glob1(abspath, "*.mfn")[0]

        try:
            instance._mf = fp.modflow.Modflow.load(
                os.path.join(abspath, name_file),
                check=True,
                forgive=True,
                model_ws=abspath,
                verbose=True,
            )

            sys.stdout = orig_stdout
            f.close()

            return instance

        except:
            sys.stdout = orig_stdout
            f.close()
            raise

    @staticmethod
    def load_with_crs(path, xll, yll, epsg=4326, rot=None):
        try:
            instance = InowasModflowReadAdapter.load(path)

            # Transform to UTM with WGS84 as intermediate
            tf = Transformer.from_crs(epsg, 4326, always_xy=True, skip_equivalent=True)
            wgs84_xll, wgs84_yll = tf.transform(xll, yll)

            # Calculate the UTM-EPSG to use with pyproj
            _, _, zone_number, _ = utm.from_latlon(wgs84_yll, wgs84_xll)
            utm_epsg = f"326{zone_number:02d}" if wgs84_yll >= 0 else f"327{zone_number:02d}"

            tf = Transformer.from_crs(4326, int(utm_epsg), always_xy=True, skip_equivalent=True)
            xoff, yoff = tf.transform(wgs84_xll, wgs84_yll)

            modelgrid = instance._mf.modelgrid
            updated_modelgrid = StructuredGrid(
                modelgrid.delc,
                modelgrid.delr,
                modelgrid.top,
                modelgrid.botm,
                modelgrid.idomain,
                modelgrid.lenuni,
                epsg=utm_epsg,
                xoff=xoff,
                yoff=yoff,
                angrot=rot if not None else modelgrid.angrot
            )

            instance._mf.modelgrid = updated_modelgrid
            return instance

        except:
            raise

    def __init__(self):
        pass

    @property
    def modelgrid(self):
        if not isinstance(self._mf, fp.modflow.Modflow):
            raise FileNotFoundError('Model not loaded')

        return self._mf.modelgrid

    @property
    def modeltime(self):
        if not isinstance(self._mf, fp.modflow.Modflow):
            raise FileNotFoundError('Model not loaded')

        return self._mf.modeltime

    def get_ibound(self):
        if not isinstance(self._mf, fp.modflow.Modflow):
            raise FileNotFoundError('Model not loaded')

        bas_package = None
        package_list = self._mf.get_package_list()

        if 'BAS' in package_list:
            bas_package = self._mf.get_package('BAS')

        if 'BAS6' in package_list:
            bas_package = self._mf.get_package('BAS6')

        if not isinstance(bas_package, fp.modflow.ModflowBas):
            raise Exception('Bas package could not be loaded.')

        return bas_package.ibound.array

    @staticmethod
    def wgs82ToUtm(x, y):
        easting, northing, zone_number, zone_letter = utm.from_latlon(y, x)
        return easting, northing, zone_number, zone_letter

    @staticmethod
    def utmToWgs82XY(easting, northing, zone_number, zone_letter):
        latitude, longitude = utm.to_latlon(easting, northing, zone_number, zone_letter)
        return longitude, latitude

    @staticmethod
    def transform(tf: Transformer, cell):
        return tf.transform(cell[0], cell[1])

    def get_cell_center(self, c):
        grid: StructuredGrid = self.modelgrid
        xcz = grid.xcellcenters
        ycz = grid.ycellcenters
        nx, ny = int(c[0]), int(c[1])
        return [xcz[ny][nx], ycz[ny][nx]]

    def model_geometry(self, target_epsg=4326, layer=0):

        i_bound = self.get_ibound()
        if layer >= len(i_bound):
            raise Exception('Layer with key ' + str(layer) + 'not found. Max: ' + str(len(i_bound)))

        layer = i_bound[layer]
        mask = np.array(np.ma.masked_values(layer, 1, shrink=False), dtype=bool)
        mpoly_cells = []
        for vec in rasterio.features.shapes(layer, mask=mask):
            mpoly_cells.append(geojson.Polygon(vec[0]["coordinates"]))

        mpoly_cells = mpoly_cells[0]
        mpoly_coordinates_utm = geojson.utils.map_tuples(lambda c: self.get_cell_center(c), mpoly_cells)

        tf = Transformer.from_crs(int(self.modelgrid.epsg), int(target_epsg), always_xy=True, skip_equivalent=True)
        mpoly_coordinates_target = geojson.utils.map_tuples(lambda c: self.transform(tf, c), mpoly_coordinates_utm)

        # noinspection PyTypeChecker
        polygon = geojson.Polygon(mpoly_coordinates_target['coordinates'])
        return polygon

    def wel_boundaries(self, target_epsg=4326):
        return fpa.WelAdapter.generate_import(self._mf, target_epsg=target_epsg)

    def model_grid_size(self):
        if not isinstance(self._mf, fp.modflow.Modflow):
            raise FileNotFoundError('Model not loaded.')

        nrow, ncol, nlay, nper = self._mf.get_nrow_ncol_nlay_nper()

        return {
            'n_x': ncol,
            'n_y': nrow
        }

    def model_stress_periods(self, start_datetime=None):
        if not isinstance(self._mf, fp.modflow.Modflow):
            raise FileNotFoundError('Model not loaded.')

        if not isinstance(self._mf.dis, fp.modflow.ModflowDis):
            raise FileNotFoundError('Dis-Package not loaded.')

        time_unit = self._mf.dis.itmuni

        if time_unit != 4:
            raise InvalidModelUnitException('The time unit is required to be in days (4)')

        from datetime import datetime, timedelta
        mt = self._mf.modeltime

        if start_datetime is None:
            start_datetime = datetime.strptime(mt.start_datetime, '%m-%d-%Y')

        if not isinstance(start_datetime, datetime):
            raise InvalidModelUnitException('DateTime has to be None or instance od datetime.')

        end_datetime = start_datetime + timedelta(days=sum(mt.perlen.tolist()))

        stressperiods = []
        for sp_idx in range(0, mt.nper):
            start_date_time = start_datetime + timedelta(days=sum(mt.perlen.tolist()[0:sp_idx]))
            stressperiods.append({
                'start_date_time': str(start_date_time.date()),
                'nstp': mt.nstp.tolist()[sp_idx],
                'tsmult': mt.tsmult.tolist()[sp_idx],
                'steady': mt.steady_state[sp_idx]
            })

        return {
            'start_date_time': str(start_datetime.date()),
            'end_date_time': str(end_datetime.date()),
            'time_unit': time_unit,
            'stressperiods': stressperiods
        }

    def model_length_unit(self):
        if not isinstance(self._mf, fp.modflow.Modflow):
            raise FileNotFoundError('Model not loaded')

        if not isinstance(self._mf.dis, fp.modflow.ModflowDis):
            raise FileNotFoundError('Dis-Package not loaded.')

        return self._mf.dis.lenuni

    def model_time_unit(self):
        if not isinstance(self._mf, fp.modflow.Modflow):
            raise FileNotFoundError('Model not loaded')

        if not isinstance(self._mf.dis, fp.modflow.ModflowDis):
            raise FileNotFoundError('Dis-Package not loaded.')

        return self._mf.dis.itmuni



import numpy as np
import flopy.mt3d as mt


class DspAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mt):
        content = self.merge()
        return mt.Mt3dDsp(
            _mt,
            **content
        )

    @staticmethod
    def default():
        default = {
            "al": 0.01,
            "trpt": 0.1,
            "trpv": 0.01,
            "dmcoef": 1e-09,
            "extension": 'dsp',
            "multiDiff": False,
            "unitnumber": None,
            "filenames": None
        }
        return default

    @staticmethod
    def read_package(package):
        content = {
            "al": package.al.array.tolist(),
            "trpt": np.reshape(package.trpt.array, (len(package.trpt.array),)).tolist(),
            "trpv": np.reshape(package.trpv.array, (len(package.trpv.array),)).tolist(),
            # "dmcoef": package.dmcoef.array.tolist(),
            "extension": package.extension[0],
            "multiDiff": package.multiDiff,
            "unitnumber": package.unit_number[0],
            # "filenames": package.filenames
        }
        return content



import flopy.modflow as mf
import geojson
import numpy as np


class WelAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            if key == 'stress_period_data':
                default[key] = self.to_dict(self._data[key])
                continue

            default[key] = self._data[key]
        return default

    def to_dict(self, data):
        if type(data) == list:
            spd_dict = {}
            for stress_period, record in enumerate(data):
                spd_dict[stress_period] = record
            return spd_dict
        return data

    def get_package(self, _mf):
        content = self.merge()
        return mf.ModflowWel(
            _mf,
            **content
        )

    @staticmethod
    def default():
        default = {
            "ipakcb": 0,
            "stress_period_data": None,
            "dtype": None,
            "extension": 'wel',
            "unitnumber": 20,
            "options": None
        }

        return default

    @staticmethod
    def read_package(package):
        content = {
            "ipakcb": package.ipakcb,
            # stress period data values translated to list of lists to be json serializable
            "stress_period_data": {k: [list(i) for i in v] for k, v in package.stress_period_data.data.items()},
            # "dtype": package.dtype,
            "extension": package.extension[0],
            "unitnumber": package.unit_number[0],
            # options is None if options list is empty:
            "options": package.options if package.options else None
        }
        return content

    @staticmethod
    def generate_import(model: mf.Modflow, target_epsg=4326):
        if not isinstance(model, mf.Modflow):
            raise FileNotFoundError('Model not loaded')

        def get_cell_center(grid, c):
            xcz = grid.xcellcenters
            ycz = grid.ycellcenters
            nx, ny = int(c[0]), int(c[1])

            return xcz[ny][nx], ycz[ny][nx]

        default = 0
        try:
            wel: mf.ModflowWel = model.wel
            flux = np.array(wel.stress_period_data.array["flux"])
            flux_cells = np.argwhere(~np.isnan(flux))

            well_cells = []
            for cell in flux_cells:
                sp, l, r, c = cell
                if [l, r, c] not in well_cells:
                    well_cells.append([l, r, c])

            from pyproj import Transformer
            tf = Transformer.from_crs(int(model.modelgrid.epsg), int(target_epsg), always_xy=True, skip_equivalent=True)

            wel_boundaries = []
            for idx, cell in enumerate(well_cells):
                l, r, c = cell
                center_x, center_y = get_cell_center(model.modelgrid, [c, r])
                center = tf.transform(center_x, center_y)

                sp_values = []
                for spd in flux:
                    value = spd[l][r][c]
                    if ~np.isnan(value):
                        sp_values.append(value)
                        continue

                    sp_values.append(default)

                wel_boundaries.append({
                    'type': 'wel',
                    'name': 'Well ' + str(idx + 1),
                    'geometry': geojson.Point(center),
                    'layers': [l],
                    'sp_values': sp_values,
                    'cells': [[c, r]]
                })

            return wel_boundaries

        except AttributeError:
            return []



import flopy.modflow as mf


class FhbAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]

        return default

    def get_package(self, _mf):
        content = self.merge()
        return mf.ModflowFhb(
            _mf,
            **content
        )

    @staticmethod
    def default():
        default = {
            "nbdtim": 1,
            "nflw": 0,
            "nhed": 0,
            "ifhbss": 0,
            "ipakcb": None,
            "nfhbx1": 0,
            "nfhbx2": 0,
            "ifhbpt": 0,
            "bdtimecnstm": 1.0,
            "bdtime": [0.0],
            "cnstm5": 1.0,
            "ds5": None,
            "cnstm7": 1.0,
            "ds7": None,
            "extension": 'fhb',
            "unitnumber": None,
            "filenames": None
        }

        return default

    @staticmethod
    def read_package(package):
        content = {
            "nbdtim": package.nbdtim,
            "nflw": package.nflw,
            "nhed": package.nhed,
            "ifhbss": package.ifhbss,
            "nfhbx1": package.nfhbx1,
            "nfhbx2": package.nfhbx2,
            "ifhbpt": package.ifhbpt,
            "bdtimecnstm": package.bdtimecnstm,
            "bdtime": package.bdtime,
            "cnstm5": package.cnstm5,
            "ds5": package.ds5,
            "cnstm7": package.cnstm7,
            "ds7": package.ds7,
            "extension": package.extension,
            "unitnumber": package.unitnumber,
            "filenames": package.filenames,
        }
        return content



import flopy.mt3d as mt


class LktAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mt):
        content = self.merge()
        return mt.Mt3dLkt(
            _mt,
            **content
        )

    @staticmethod
    def default():
        default = {
            "nlkinit": 0,
            "mxlkbc": 0,
            "icbclk": None,
            "ietlak": 0,
            "coldlak": 0.0,
            "lk_stress_period_data": None,
            "dtype": None,
            "extension": 'lkt',
            "unitnumber": None,
            "filenames": None
        }
        return default

    @staticmethod
    def read_package(package):
        content = {
            "nlkinit": package.nlkinit,
            "mxlkbc": package.mxlkbc,
            "icbclk": package.icbclk,
            "ietlak": package.ietlak,
            "coldlak": package.coldlak,
            "lk_stress_period_data": package.lk_stress_period_data,
            "dtype": package.dtype,
            "extension": package.extension,
            "unitnumber": package.unitnumber,
            "filenames": package.filenames
        }
        return content



import json
import numpy as np
import os
from scipy import stats
from sklearn.metrics import r2_score


class HobStatistics:

    def __init__(self, model_ws, name):
        self._model_ws = model_ws
        self._name = name
        self._input_file = os.path.join(model_ws, name) + '.hob.out'
        self._output_file = os.path.join(model_ws, name) + '.hob.stat'

    def write_files(self):
        with open(self._output_file, 'w') as outfile:
            try:
                result = self.calculate()
            except:
                result = {"error": 'Error in Hob-Calculation.'}
            finally:
                json.dump(result, outfile)

    @staticmethod
    def calculate_npf(x, n):
        a = 0.5
        if x < 11:
            a = 3 / 8

        return stats.norm.ppf((x - a) / (n + 1 - 2 * a))

    def calculate(self):
        if not os.path.isfile(self._input_file):
            return {"error": 'File ' + self._input_file + ' not found.'}

        f = open(self._input_file)

        header = False

        names = []
        observed = []
        simulated = []

        for line in f:
            if line.startswith('#'):
                continue

            if not header:
                header = line.split('"')[1::2]
                continue

            values = line.split()
            simulated.append(float(values[0]))
            observed.append(float(values[1]))
            names.append('_'.join(values[2].split('_')[:-1]))

        simulated = np.array(simulated)
        observed = np.array(observed)

        # Write to statistics object
        statistics = dict(
            observed=list(observed),
            simulated=list(simulated),
            n=len(observed),
            rMax=np.max(np.abs(simulated - observed)),
            rMin=np.min(np.abs(simulated - observed)),
            rMean=np.mean(simulated - observed),
            absRMean=np.mean(np.abs(simulated - observed)),
            sse=stats.sem(simulated - observed),
            rmse=np.sqrt(((simulated - observed) ** 2).mean()),
            R=stats.pearsonr(observed, simulated)[0],
            R2=r2_score(observed, simulated)
        )

        statistics["nrmse"] = statistics["rmse"] / (np.max(observed) - np.min(observed))

        # Plot simulated vs. observed values
        statistics["Z"] = 1.96
        statistics["stdObserved"] = np.std(observed)
        statistics["deltaStd"] = statistics["Z"] * statistics["stdObserved"] / np.sqrt(statistics["n"])

        # Plot (weighted) residuals vs. simulated heads
        statistics["weightedResiduals"] = list(simulated - observed)
        statistics["linRegressSW"] = stats.linregress(simulated, statistics["weightedResiduals"])

        # Check for Normal Distribution¶
        statistics["rankedResiduals"] = list(np.sort(simulated - observed))

        n = statistics["n"]
        npf = np.linspace(1, n, num=n)
        npf = list(map(lambda x, n: self.calculate_npf(x, n), npf, np.ones(n) * n))

        statistics["npf"] = npf
        statistics["linRegressRN"] = stats.linregress(statistics["rankedResiduals"], statistics["npf"])

        return statistics



import flopy.mt3d as mt


class GcgAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mt):
        content = self.merge()
        return mt.Mt3dGcg(
            _mt,
            **content
        )

    @staticmethod
    def default():
        default = {
            "mxiter": 1,
            "iter1": 50,
            "isolve": 3,
            "ncrs": 0,
            "accl": 1,
            "cclose": 1e-05,
            "iprgcg": 0,
            "extension": 'gcg',
            "unitnumber": None
        }
        return default

    @staticmethod
    def read_package(package):
        content = {
            "mxiter": package.mxiter,
            "iter1": package.iter1,
            "isolve": package.isolve,
            "ncrs": package.ncrs,
            "accl": package.accl,
            "cclose": package.cclose,
            "iprgcg": package.iprgcg,
            "extension": package.extension[0],
            "unitnumber": package.unit_number[0]
        }
        return content



import flopy.modflow as mf


class BasAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mf):
        content = self.merge()
        return mf.ModflowBas(
            _mf,
            **content
        )

    @staticmethod
    def default():
        default = {
            "ibound": 1,
            "strt": 1.0,
            "ifrefm": True,
            "ixsec": False,
            "ichflg": False,
            "stoper": None,
            "hnoflo": -999.99,
            "extension": 'bas',
            "unitnumber": 13
        }
        return default

    @staticmethod
    def read_package(package):
        content = {
            "ibound": package.ibound.array.tolist(),
            "strt": package.strt.array.tolist(),
            "ifrefm": package.ifrefm,
            "ixsec": package.ixsec,
            "ichflg": package.ichflg,
            "stoper": package.stoper,
            "hnoflo": package.hnoflo,
            "extension": package.extension[0],
            "unitnumber": package.unit_number[0]
        }
        return content



import flopy.modflow as mf


class NwtAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mf):
        content = self.merge()
        return mf.ModflowNwt(
            _mf,
            **content
        )

    @staticmethod
    def default():
        default = {
            "headtol": 1E-2,
            "fluxtol": 500,
            "maxiterout": 100,
            "thickfact": 1E-5,
            "linmeth": 1,
            "iprnwt": 0,
            "ibotav": 0,
            "options": 'COMPLEX',
            "Continue": False,
            "dbdtheta": 0.4,
            "dbdkappa": 1.e-5,
            "dbdgamma": 0.,
            "momfact": 0.1,
            "backflag": 1,
            "maxbackiter": 50,
            "backtol": 1.1,
            "backreduce": 0.70,
            "maxitinner": 50,
            "ilumethod": 2,
            "levfill": 5,
            "stoptol": 1.e-10,
            "msdr": 15,
            "iacl": 2,
            "norder": 1,
            "level": 5,
            "north": 7,
            "iredsys": 0,
            "rrctols": 0.0,
            "idroptol": 1,
            "epsrn": 1.e-4,
            "hclosexmd": 1e-4,
            "mxiterxmd": 50,
            "extension": 'nwt',
            "unitnumber": 32
        }

        return default

    @staticmethod
    def read_package(package):
        content = {
            "fluxtol": package.fluxtol,
            "maxiterout": package.maxiterout,
            "thickfact": package.thickfact,
            "linmeth": package.linmeth,
            "iprnwt": package.iprnwt,
            # "ibotavg": package.ibotavg,
            "options": package.options,
            # "continue": package.continue,
            "dbdtheta": package.dbdtheta,
            "dbdkappa": package.dbdkappa,
            "dbdgamma": package.dbdgamma,
            "momfact": package.momfact,
            "backflag": package.backflag,
            "maxbackiter": package.maxbackiter,
            "backtol": package.backtol,
            "backreduce": package.backreduce,
            "maxitinner": package.maxitinner,
            "ilumethod": package.ilumethod,
            "levfill": package.levfill,
            "stoptol": package.stoptol,
            "msdr": package.msdr,
            "iacl": package.iacl,
            "norder": package.norder,
            "level": package.level,
            "north": package.north,
            "iredsys": package.iredsys,
            "rrctols": package.rrctols,
            "idroptol": package.idroptol,
            "epsrn": package.epsrn,
            "hclosexmd": package.hclosexmd,
            "mxiterxmd": package.mxiterxmd,
            "extension": package.extension[0],
            "unitnumber": package.unit_number[0]
        }
        return content



import unittest
from ...Read import InowasModflowReadAdapter


class InowasModflowReadAdapterTest(unittest.TestCase):
    def it_can_be_instantiated_test(self):
        instance = InowasModflowReadAdapter()
        self.assertIsInstance(instance, InowasModflowReadAdapter)

    def it_throws_exception_if_path_is_wrong_test(self):
        with self.assertRaises(FileNotFoundError) as context:
            InowasModflowReadAdapter.load('abc')
        self.assertEqual('Path not found: abc', str(context.exception))

    def it_throws_exception_if_path_does_not_contain_name_file_test(self):
        with self.assertRaises(FileNotFoundError) as context:
            InowasModflowReadAdapter.load('./FlopyAdapter/test/Read/data/emptyFolder')
        self.assertEqual('Modflow name file with ending .nam or .mfn not found', str(context.exception))

    def it_loads_the_model_correctly_test(self):
        instance = InowasModflowReadAdapter.load('./FlopyAdapter/test/Read/data/test_example_1')
        self.assertIsInstance(instance, InowasModflowReadAdapter)

    def it_loads_the_model_correctly_with_crs_test(self):
        instance = InowasModflowReadAdapter.load_with_crs(
            './FlopyAdapter/test/Read/data/test_example_1',
            279972.0566, 9099724.9436, 31985, -15.5
        )

        self.assertIsInstance(instance, InowasModflowReadAdapter)

        from flopy.discretization import StructuredGrid
        mg = instance.modelgrid
        self.assertIsInstance(mg, StructuredGrid)
        self.assertEqual(mg.epsg, '32725')
        self.assertEqual(round(mg.xoffset), 279972)
        self.assertEqual(round(mg.yoffset), 9099725)
        self.assertEqual(mg.angrot, -15.5)

    def it_converts_wgs84_to_utm_correctly_test(self):
        lat = 50.966319
        long = 13.923273
        easting, northing, zone_number, zone_letter = InowasModflowReadAdapter.wgs82ToUtm(long, lat)
        self.assertEqual(round(easting), 424393)
        self.assertEqual(round(northing), 5646631)
        self.assertEqual(zone_number, 33)
        self.assertEqual(zone_letter, 'U')

    def it_converts_utm_to_wgs84_correctly_test(self):
        easting, northing, zone_number, zone_letter = 424393, 5646631, 33, 'U'
        long, lat = InowasModflowReadAdapter.utmToWgs82XY(easting, northing, zone_number, zone_letter)
        self.assertEqual(round(long, 5), 13.92328)
        self.assertEqual(round(lat, 5), 50.96632)

    def it_returns_a_model_geometry_correctly_test(self):
        instance = InowasModflowReadAdapter.load_with_crs(
            './FlopyAdapter/test/Read/data/test_example_1',
            279972.0566, 9099724.9436, 31985, -15.5
        )

        geometry = instance.model_geometry(4326, 0)
        self.assertEqual(geometry["type"], 'Polygon')
        self.assertEqual(len(geometry["coordinates"][0]), 520)
        self.assertEqual(geometry["coordinates"][0][0], [-34.874831, -8.073991])

    def it_returns_model_gid_size_test(self):
        instance = InowasModflowReadAdapter.load('./FlopyAdapter/test/Read/data/test_example_1')
        self.assertIsInstance(instance, InowasModflowReadAdapter)
        grid_size = instance.model_grid_size()
        self.assertEqual(grid_size, {
            'n_x': 227,
            'n_y': 221,
        })

    def it_returns_model_stressperiods_test(self):
        instance = InowasModflowReadAdapter.load('./FlopyAdapter/test/Read/data/test_example_1')
        self.assertIsInstance(instance, InowasModflowReadAdapter)
        stress_periods = instance.model_stress_periods()
        expected = {
            'start_date_time': '1970-01-01',
            'end_date_time': '1970-02-06',
            'time_unit': 4,
            'stressperiods': [
                {
                    'start_date_time': '1970-01-01',
                    'nstp': 1,
                    'tsmult': 1.0,
                    'steady': True,
                },
                {
                    'start_date_time': '1970-01-02',
                    'nstp': 5,
                    'tsmult': 1.0,
                    'steady': False,
                },
                {
                    'start_date_time': '1970-01-12',
                    'nstp': 10,
                    'tsmult': 1.5,
                    'steady': False,
                }
            ]
        }
        self.assertEqual(stress_periods, expected)

    def it_returns_model_length_unit_test(self):
        instance = InowasModflowReadAdapter.load('./FlopyAdapter/test/Read/data/test_example_1')
        self.assertIsInstance(instance, InowasModflowReadAdapter)
        length_unit = instance.model_length_unit()
        self.assertEqual(length_unit, 2)

    def it_returns_model_time_unit_test(self):
        instance = InowasModflowReadAdapter.load('./FlopyAdapter/test/Read/data/test_example_1')
        self.assertIsInstance(instance, InowasModflowReadAdapter)
        time_unit = instance.model_time_unit()
        self.assertEqual(time_unit, 4)

    def it_returns_wel_boundaries_of_example_1_test(self):
        instance = InowasModflowReadAdapter.load_with_crs(
            './FlopyAdapter/test/Read/data/test_example_1',
            279972.0566, 9099724.9436, 31985, -15.5
        )
        self.assertIsInstance(instance, InowasModflowReadAdapter)
        wel_boundaries = instance.wel_boundaries(target_epsg=4326)
        self.assertEqual(len(wel_boundaries), 93)
        print(wel_boundaries[0])
        self.assertEqual(wel_boundaries[0],
                         {
                             'type': 'wel',
                             'name': 'Well 1',
                             'geometry': {"coordinates": [-34.879083, -8.084035], "type": "Point"},
                             'layers': [0],
                             'sp_values': [-2039.0, -2039.0, -2039.0], 'cells': [[217, 31]]
                         }
                         )

    def it_returns_wel_boundaries_of_example_2_test(self):
        instance = InowasModflowReadAdapter.load_with_crs(
            './FlopyAdapter/test/Read/data/test_example_2',
            0, 0, 4326, 0
        )
        self.assertIsInstance(instance, InowasModflowReadAdapter)
        wel_boundaries = instance.wel_boundaries(target_epsg=4326)
        self.assertEqual(len(wel_boundaries), 6)
        self.assertEqual(wel_boundaries, [
            {
                'type': 'wel',
                'name': 'Well 1',
                'geometry': {"coordinates": [0.013461, 0.040657], "type": "Point"},
                'layers': [0],
                'sp_values': [0, -5000.0, -5000.0],
                'cells': [[1, 1]]
            }, {
                'type': 'wel',
                'name': 'Well 2',
                'geometry': {"coordinates": [0.022435, 0.040658], "type": "Point"},
                'layers': [0],
                'sp_values': [0, -5000.0, -5000.0],
                'cells': [[2, 1]]
            },
            {
                'type': 'wel',
                'name': 'Well 3',
                'geometry': {"coordinates": [0.058334, 0.040659], "type": "Point"},
                'layers': [0],
                'sp_values': [0, -10000.0, -10000.0],
                'cells': [[6, 1]]
            },
            {
                'type': 'wel',
                'name': 'Well 4',
                'geometry': {"coordinates": [0.085259, 0.04066], "type": "Point"},
                'layers': [0],
                'sp_values': [0, -5000.0, -5000.0],
                'cells': [[9, 1]]
            },
            {
                'type': 'wel',
                'name': 'Well 5',
                'geometry': {"coordinates": [0.013461, 0.022587], "type": "Point"},
                'layers': [0],
                'sp_values': [0, -5000.0, -5000.0],
                'cells': [[1, 3]]
            },
            {
                'type': 'wel',
                'name': 'Well 6',
                'geometry': {"coordinates": [0.040385, 0.013553], "type": "Point"},
                'layers': [0],
                'sp_values': [0, -5000.0, -5000.0],
                'cells': [[4, 4]]
            }])


if __name__ == "__main__":
    unittest.main()



import flopy.mt3d as mt


class AdvAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            if not key.startswith('_'):
                default[key] = self._data[key]
        return default

    def get_package(self, _mt):
        content = self.merge()
        return mt.Mt3dAdv(
            _mt,
            **content
        )

    @staticmethod
    def default():
        return {
            "mixelm": 3,
            "percel": 0.75,
            "mxpart": 800000,
            "nadvfd": 1,
            "itrack": 3,
            "wd": 0.5,
            "dceps": 1e-05,
            "nplane": 2,
            "npl": 10,
            "nph": 40,
            "npmin": 5,
            "npmax": 80,
            "nlsink": 0,
            "npsink": 15,
            "dchmoc": 0.0001,
            "extension": 'adv',
            "unitnumber": None,
            "filenames": None
        }

    @staticmethod
    def read_package(package):
        return {
            "mixelm": package.mixelm,
            "percel": package.percel,
            "mxpart": package.mxpart,
            "nadvfd": package.nadvfd,
            "itrack": package.itrack,  # None
            "wd": package.wd,  # None
            "dceps": package.dceps,  # None
            "nplane": package.nplane,  # None
            "npl": package.npl,  # None
            "nph": package.nph,  # None
            "npmin": package.npmin,  # None
            "npmax": package.npmax,  # None
            "nlsink": package.nlsink,  # None
            "npsink": package.npsink,  # None
            "dchmoc": package.dchmoc,  # None
            "extension": package.extension[0],
            "unitnumber": package.unit_number[0],
            # "filenames": package.filenames
        }



"""
This module is an intermediate layer between flopy version 3.2
and the inowas-modflow-configuration format.

Author: Ralf Junghanns
EMail: ralf.junghanns@gmail.com
"""

from FlopyAdapter.MfPackages.BasAdapter import BasAdapter
from FlopyAdapter.MfPackages.ChdAdapter import ChdAdapter
from FlopyAdapter.MfPackages.DisAdapter import DisAdapter
from FlopyAdapter.MfPackages.DrnAdapter import DrnAdapter
from FlopyAdapter.MfPackages.FhbAdapter import FhbAdapter
from FlopyAdapter.MfPackages.GhbAdapter import GhbAdapter
from FlopyAdapter.MfPackages.HobAdapter import HobAdapter
from FlopyAdapter.Statistics.HobStatistics import HobStatistics
from FlopyAdapter.MfPackages.LpfAdapter import LpfAdapter
from FlopyAdapter.MfPackages.MfAdapter import MfAdapter
from FlopyAdapter.MfPackages.NwtAdapter import NwtAdapter
from FlopyAdapter.MfPackages.OcAdapter import OcAdapter
from FlopyAdapter.MfPackages.PcgAdapter import PcgAdapter
from FlopyAdapter.MfPackages.RchAdapter import RchAdapter
from FlopyAdapter.MfPackages.EvtAdapter import EvtAdapter
from FlopyAdapter.MfPackages.RivAdapter import RivAdapter
from FlopyAdapter.Read.ReadBudget import ReadBudget
from FlopyAdapter.Read.ReadConcentration import ReadConcentration
from FlopyAdapter.Read.ReadDrawdown import ReadDrawdown
from FlopyAdapter.Read.ReadHead import ReadHead
from FlopyAdapter.MfPackages.UpwAdapter import UpwAdapter
from FlopyAdapter.MfPackages.WelAdapter import WelAdapter
from FlopyAdapter.MfPackages.LmtAdapter import LmtAdapter

from FlopyAdapter.MpPackages.MpAdapter import MpAdapter
from FlopyAdapter.MpPackages.BasAdapter import BasAdapter
from FlopyAdapter.MpPackages.SimAdapter import SimAdapter

from FlopyAdapter.MtPackages.MtAdapter import MtAdapter
from FlopyAdapter.MtPackages.AdvAdapter import AdvAdapter
from FlopyAdapter.MtPackages.BtnAdapter import BtnAdapter
from FlopyAdapter.MtPackages.DspAdapter import DspAdapter
from FlopyAdapter.MtPackages.GcgAdapter import GcgAdapter
from FlopyAdapter.MtPackages.LktAdapter import LktAdapter
from FlopyAdapter.MtPackages.PhcAdapter import PhcAdapter
from FlopyAdapter.MtPackages.RctAdapter import RctAdapter
from FlopyAdapter.MtPackages.SftAdapter import SftAdapter
from FlopyAdapter.MtPackages.SsmAdapter import SsmAdapter
from FlopyAdapter.MtPackages.TobAdapter import TobAdapter
from FlopyAdapter.MtPackages.UztAdapter import UztAdapter
from FlopyAdapter.SwtPackages.SwtAdapter import SwtAdapter
from FlopyAdapter.SwtPackages.VdfAdapter import VdfAdapter
from FlopyAdapter.SwtPackages.VscAdapter import VscAdapter


class InowasFlopyCalculationAdapter:
    """The Flopy Class"""

    _version = None
    _uuid = None

    _model = None

    _report = ''
    _success = False

    mf_package_order = [
        'mf', 'dis', 'bas', 'bas6',
        'chd', 'evt', 'drn', 'fhb', 'ghb', 'hob', 'rch', 'riv', 'wel',
        'lpf', 'upw', 'pcg', 'nwt', 'oc', 'lmt', 'lmt6'
    ]

    mt_package_order = [
        'mt', 'btn', 'adv', 'dsp', 'gcg', 'ssm', 'lkt',
        'phc', 'rct', 'sft', 'tob', 'uzt'
    ]

    swt_package_order = [
        # Modflow
        'swt', 'dis', 'bas', 'bas6', 'riv', 'wel', 'rch', 'chd', 'ghb', 'hob',
        'lpf', 'upw', 'pcg', 'nwt', 'oc', 'lmt', 'lmt6',
        # Mt3D
        'btn', 'adv', 'dsp', 'gcg', 'ssm', 'lkt', 'phc', 'rct', 'sft', 'tob', 'uzt',
        # Seawat                                         
        'vdf', 'vsc'
    ]

    mp_package_order = [
        'mp', 'bas', 'sim'
    ]

    def __init__(self, version, data, uuid):
        self._mf_data = data.get('mf')
        self._mp_data = data.get('mp')
        self._mt_data = data.get('mt')
        self._swt_data = data.get('swt')
        self._version = version
        self._uuid = uuid

        # Model calculation if Seawat is enabled
        if self._swt_data is not None:
            package_data = {
                **self._mf_data,
                **self._mt_data,
                **self._swt_data,
                'packages': self._mf_data['packages'] + self._mt_data['packages'] + self._swt_data['packages']
            }

            package_content = self.read_packages(package_data)
            self.create_model(self.swt_package_order, package_content)
            self.write_input_model(self._model)
            self._success, report = self.run_model(self._model, model_type='swt')
            self._report += report

            if 'hob' in self._mf_data['packages']:
                print('Calculate hob-statistics and write to file %s.hob.stat' % uuid)
                self.run_hob_statistics(self._model)

            return

        # Normal Modflow calculation
        if self._mf_data is not None:
            package_content = self.read_packages(self._mf_data)
            self.create_model(self.mf_package_order, package_content)
            self.write_input_model(self._model)
            self.success, report = self.run_model(self._model, model_type='mf')
            self._report += report

            if 'hob' in self._mf_data['packages']:
                print('Calculate hob-statistics and write to file %s.hob.stat' % uuid)
                self.run_hob_statistics(self._model)

            # Mt3d calculation
            if self._mt_data is not None:
                package_content = self.read_packages(self._mt_data)
                self.create_model(self.mt_package_order, package_content)
                self.write_input_model(self._model)
                self._success, report = self.run_model(self._model, model_type='mt')
                self._report += report

            # ModPath6 calculation
            if self._mp_data is not None:
                package_content = self.read_packages(self._mp_data)
                self.create_model(self.mp_package_order, package_content)
                self.write_input_model(self._model)
                self._success, report = self.run_model(self._model, model_type='mp')
                self._report += report

    @staticmethod
    def read_packages(data):
        package_content = {}
        for package in data['packages']:
            print('Read Flopy package data: %s' % package)
            package_content[package.lower()] = data[package]
        return package_content

    def create_model(self, package_order, package_content):
        for package in package_order:
            if package in package_content:
                print('Create Flopy Package: %s' % package)
                self.create_package(package, package_content[package])

    @staticmethod
    def write_input_model(model):
        print('Write input files.')
        model.write_input()

    @staticmethod
    def run_model(model, model_type):
        normal_msg = 'normal termination'
        if model_type == 'mt':
            normal_msg = 'Program completed'

        print('Run model.')
        print('Model nam-file: %s.' % model.namefile)
        print('Model executable: %s.' % model.exe_name)
        success, report = model.run_model(report=True, silent=True, normal_msg=normal_msg)
        return success, ' \n'.join(str(e) for e in report)

    @staticmethod
    def run_hob_statistics(model):
        model_ws = model.model_ws
        name = model.name

        print('Calculate hob-statistics for model %s' % name)
        HobStatistics(model_ws, name).write_to_file()

    def check_model(self):
        if self._model is not None:
            self._model.check()
        if self._model is not None:
            self._model.check()

    def create_package(self, name, content):

        # Modflow packages
        if name == 'mf':
            self._model = MfAdapter(content).get_package()
        if name == 'dis':
            DisAdapter(content).get_package(self._model)
        if name == 'drn':
            DrnAdapter(content).get_package(self._model)
        if name == 'bas' or name == 'bas6':
            BasAdapter(content).get_package(self._model)
        if name == 'lpf':
            LpfAdapter(content).get_package(self._model)
        if name == 'upw':
            UpwAdapter(content).get_package(self._model)
        if name == 'pcg':
            PcgAdapter(content).get_package(self._model)
        if name == 'nwt':
            NwtAdapter(content).get_package(self._model)
        if name == 'oc':
            OcAdapter(content).get_package(self._model)
        if name == 'riv':
            RivAdapter(content).get_package(self._model)
        if name == 'wel':
            WelAdapter(content).get_package(self._model)
        if name == 'rch':
            RchAdapter(content).get_package(self._model)
        if name == 'evt':
            EvtAdapter(content).get_package(self._model)
        if name == 'chd':
            ChdAdapter(content).get_package(self._model)
        if name == 'fhb':
            FhbAdapter(content).get_package(self._model)
        if name == 'ghb':
            GhbAdapter(content).get_package(self._model)
        if name == 'hob':
            HobAdapter(content).get_package(self._model)
        if name == 'lmt':
            LmtAdapter(content).get_package(self._model)

        # MT3D packages
        if name == 'mt':
            self._model = MtAdapter(content).get_package(self._model)
        if name == 'adv':
            AdvAdapter(content).get_package(self._model)
        if name == 'btn':
            BtnAdapter(content).get_package(self._model)
        if name == 'dsp':
            DspAdapter(content).get_package(self._model)
        if name == 'gcg':
            GcgAdapter(content).get_package(self._model)
        if name == 'lkt':
            LktAdapter(content).get_package(self._model)
        if name == 'phc':
            PhcAdapter(content).get_package(self._model)
        if name == 'rct':
            RctAdapter(content).get_package(self._model)
        if name == 'sft':
            SftAdapter(content).get_package(self._model)
        if name == 'ssm':
            SsmAdapter(content).get_package(self._model)
        if name == 'tob':
            TobAdapter(content).get_package(self._model)
        if name == 'uzt':
            UztAdapter(content).get_package(self._model)

        # ModPath packages
        if name == 'mp':
            self._model = MpAdapter(content).get_package()
        if name == 'mpbas':
            self._model = BasAdapter(content).get_package(self._model)
        if name == 'mpsim':
            self._model = SimAdapter(content).get_package(self._model)

        # Seawat packages
        if name == 'swt':
            self._model = SwtAdapter(content).get_package()
        if name == 'vdf':
            VdfAdapter(content).get_package(self._model)
        if name == 'vsc':
            VscAdapter(content).get_package(self._model)

    def response(self):
        key = 'mf'
        if 'MF' in self._mf_data:
            key = 'MF'

        budgets = ReadBudget(self._mf_data[key]['model_ws'])
        concentrations = ReadConcentration(self._mf_data[key]['model_ws'])
        drawdowns = ReadDrawdown(self._mf_data[key]['model_ws'])
        heads = ReadHead(self._mf_data[key]['model_ws'])

        return {
            'budgets': budgets.read_times(),
            'concentrations': concentrations.read_times(),
            'drawdowns': drawdowns.read_times(),
            'heads': heads.read_times(),
            'number_of_layers': heads.read_number_of_layers()
        }

    def success(self):
        return self._success

    def response_message(self):
        return self._report



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="InowasFlopyAdapter",
    version="1.1.1",
    author="Ralf Junghanns",
    author_email="ralf.junghanns@gmail.com",
    description="A FLOPY wrapper for the INOWAS-platform",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/inowas/InowasFlopyAdapter",
    packages=setuptools.find_packages(),
    install_requires=[
        'flopy==3.3.1',
        'geojson>=2.5',
        'nose>=1.3',
        'numpy>=1.9',
        'pyproj>=2.6',
        'rasterio>=1.1',
        'scipy>=1.4',
        'shapely>=1.7',
        'sklearn>=0.0',
        'setuptools>=50.0',
        'twine>=3.2',
        'utm>=0.5.0'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



import flopy.modflow as mf


class LpfAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mf):
        content = self.merge()
        return mf.ModflowLpf(
            _mf,
            **content
        )

    @staticmethod
    def default():
        default = {
            "laytyp": 0,
            "layavg": 0,
            "chani": 1.0,
            "layvka": 0,
            "laywet": 0,
            "ipakcb": 53,
            "hdry": -1e+30,
            "iwdflg": 0,
            "wetfct": 0.1,
            "iwetit": 1,
            "ihdwet": 0,
            "hk": 1.0,
            "hani": 1.0,
            "vka": 1.0,
            "ss": 1e-5,
            "sy": 0.15,
            "vkcb": 0.0,
            "wetdry": -0.01,
            "storagecoefficient": False,
            "constantcv": False,
            "thickstrt": False,
            "nocvcorrection": False,
            "novfc": False,
            "extension": 'lpf',
            "unitnumber": 15
        }

        return default

    @staticmethod
    def read_package(package):
        content = {
            "laytyp": package.laytyp.array.tolist(),
            "layavg": package.layavg.array.tolist(),
            "chani": package.chani.array.tolist(),
            "layvka": package.layvka.array.tolist(),
            "laywet": package.laywet.array.tolist(),
            "ipakcb": package.ipakcb,
            "hdry": package.hdry,
            # "iwdflg": package.iwdflg,
            "wetfct": package.wetfct,  # None
            "iwetit": package.iwetit,  # None
            "ihdwet": package.ihdwet,  # None
            "hk": package.hk.array.tolist(),
            "hani": package.hani.array.tolist(),
            "vka": package.vka.array.tolist(),
            "ss": package.ss.array.tolist(),
            "sy": package.sy.array.tolist(),
            "vkcb": package.vkcb.array.tolist(),
            "wetdry": package.wetdry.array.tolist(),
            # "storagecoefficient": package.storagecoefficient,
            # "constantcv": package.constantcv,
            # "thickstrt": package.thickstrt,
            # "nocvcorrection": package.nocvcorrection,
            # "novfc": package.novfc,
            "extension": package.extension[0],
            "unitnumber": package.unit_number[0]
        }
        return content



import flopy.modflow as mf


class PcgAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mf):
        content = self.merge()
        return mf.ModflowPcg(
            _mf,
            **content
        )

    @staticmethod
    def default():
        default = {
            "mxiter": 50,
            "iter1": 30,
            "npcond": 1,
            "hclose": 1E-5,
            "rclose": 1E-5,
            "relax": 1.0,
            "nbpol": 0,
            "iprpcg": 0,
            "mutpcg": 3,
            "damp": 1.0,
            "dampt": 1.0,
            "ihcofadd": 0,
            "extension": 'pcg',
            "unitnumber": 27
        }

        return default

    @staticmethod
    def read_package(package):
        content = {
            "mxiter": package.mxiter,
            "iter1": package.iter1,
            "npcond": package.npcond,
            "hclose": package.hclose,
            "rclose": package.rclose,
            "relax": package.relax,
            "nbpol": package.nbpol,
            "iprpcg": package.iprpcg,
            "mutpcg": package.mutpcg,
            "damp": package.damp,
            "dampt": package.dampt,
            "ihcofadd": package.ihcofadd,
            "extension": package.extension[0],
            "unitnumber": package.unit_number[0]
        }
        return content



import flopy.modflow as mf


class DisAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mf):
        content = self.merge()
        return mf.ModflowDis(
            _mf,
            **content
        )

    @staticmethod
    def default():
        default = {
            "nlay": 1,
            "nrow": 2,
            "ncol": 2,
            "nper": 1,
            "delr": 1.0,
            "delc": 1.0,
            "laycbd": 0,
            "top": 1,
            "botm": 0,
            "perlen": 1,
            "nstp": 1,
            "tsmult": 1,
            "steady": True,
            "itmuni": 4,
            "lenuni": 2,
            "extension": 'dis',
            "unitnumber": 11,
            "xul": None,
            "yul": None,
            "rotation": 0.0,
            "proj4_str": None,
            "start_datetime": None
        }
        return default

    @staticmethod
    def read_package(package):
        content = {
            "nlay": package.nlay,
            "nrow": package.nrow,
            "ncol": package.ncol,
            "nper": package.nper,
            "delr": package.delr.array.tolist(),
            "delc": package.delc.array.tolist(),
            "laycbd": package.laycbd.array.tolist(),
            "top": package.top.array.tolist(),
            "botm": package.botm.array.tolist(),
            "perlen": package.perlen.array.tolist(),
            "nstp": package.nstp.array.tolist(),
            "tsmult": package.tsmult.array.tolist(),
            "steady": package.steady.array.tolist(),
            "itmuni": package.itmuni,
            "lenuni": package.lenuni,
            "extension": package.extension[0],
            "unitnumber": package.unit_number[0],
            # "xul": package.xul,
            # "yul": package.yul,
            # "rotation": package.rotation,
            # "proj4_str": package.proj4_str,
            "start_datetime": package.start_datetime
        }
        return content



import flopy.modflow as mf


class LmtAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mf):
        content = self.merge()
        return mf.ModflowLmt(
            _mf,
            **content
        )

    @staticmethod
    def default():
        default = {
            "output_file_name": 'mt3d_link.ftl',
            "output_file_unit": 54,
            "output_file_header": 'extended',
            "output_file_format": 'unformatted',
            "extension": 'lmt6',
            "package_flows": [],
            "unitnumber": None,
            "filenames": None
        }

        return default

    @staticmethod
    def read_package(package):
        content = {
            "output_file_name": package.output_file_name,
            "output_file_unit": package.output_file_unit,
            "output_file_header": package.output_file_header,
            "output_file_format": package.output_file_format,
            "extension": package.extension[0],
            "package_flows": package.package_flows,
            "unitnumber": package.unit_number[0],
            # "filenames": None
        }
        return content



import flopy.mt3d as mt


class MtAdapter:
    _data = None

    def __init__(self, data):
        self._data = data

    def validate(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return some hints
        pass

    def is_valid(self):
        # should be implemented
        # for key in content:
        #   do something
        #   return true or false
        return True

    def merge(self):
        default = self.default()
        for key in self._data:
            default[key] = self._data[key]
        return default

    def get_package(self, _mf):
        content = self.merge()

        return mt.Mt3dms(modflowmodel=_mf, **content)

    @staticmethod
    def default():
        default = {
            "modelname": 'mt3dtest',
            "namefile_ext": 'nam',
            "ftlfilename": 'mt3d_link.ftl',
            "ftlfree": False,
            "version": 'mt3dms',
            "exe_name": 'mt3dms.exe',
            "structured": True,
            "listunit": None,
            "ftlunit": None,
            "model_ws": '.',
            "external_path": None,
            "verbose": False,
            "load": True,
            "silent": 0
        }
        return default

    @staticmethod
    def read_package(package):
        content = {
            "modelname": package.namefile.split('.')[0],
            "namefile_ext": package.namefile_ext,
            "ftlfilename": package.ftlfilename,
            "ftlfree": package.ftlfree,
            "version": package.version,
            "exe_name": package.exe_name,  # None
            "structured": package.structured,
            "listunit": package.listunit,  # None
            "ftlunit": package.ftlunit,
            "model_ws": package.model_ws,
            "external_path": package.external_path,  # None
            "verbose": package.verbose,
            "load": package.load,
            "silent": package.silent
        }
        return content
